using System;
using System.Reflection;
using System.Text;
using HarmonyLib;
using Vintagestory.API.Client;
using Vintagestory.API.Common;
using Vintagestory.API.Datastructures;
using Vintagestory.GameContent;

namespace Autoloot {
	public class AutolootConfig {
		public bool log = true;
		public bool sound = true;
	}

	public class AutolootSystem: ModSystem {
		public const string configFilename = "autoloot.json";
		protected const string harmonyId = "autoloot";

		public const string wordFalse = "off";
		public const string wordTrue = "on";
		public static readonly string[] binaryWords = new string[] { wordFalse, wordTrue };
		
		public static ICoreClientAPI capi;
		public static AutolootConfig config;
		public static StringBuilder sb;
		public static AssetLocation sfx;

		protected static GuiDialogCreatureContents dialog;
		protected static InventoryGeneric inventory;
		protected static long entityId;
		protected static bool waiting = false;

		protected Harmony harmony;

		public override bool ShouldLoad(EnumAppSide side) {
			return side == EnumAppSide.Client;
		}

		public override void StartClientSide(ICoreClientAPI capi) {
			base.StartClientSide(capi);
			AutolootSystem.capi = capi;

			// Load or create config file.
			try {
				AutolootSystem.config = capi.LoadModConfig<AutolootConfig>(configFilename);
			} catch(Exception) {}
			if(AutolootSystem.config != null) {
				// Save config file in case of missing fields.
				capi.StoreModConfig<AutolootConfig>(AutolootSystem.config, configFilename);
			} else {
				// Create new config file.
				AutolootSystem.config = new AutolootConfig();
				capi.StoreModConfig<AutolootConfig>(AutolootSystem.config, configFilename);
			}

			AutolootSystem.sb = new StringBuilder();
			AutolootSystem.sfx = new AssetLocation("autoloot", "sounds/loot");

			capi.ChatCommands.Create("autoloot")
					.BeginSubCommand("log")
						.WithDescription("Enable or disable chat logging.")
						.WithArgs(capi.ChatCommands.Parsers.WordRange("off/on", AutolootSystem.binaryWords))
						.HandleWith(this.cmdLog)
					.EndSubCommand()
					.BeginSubCommand("sound")
						.WithDescription("Enable or disable sound effect.")
						.WithArgs(capi.ChatCommands.Parsers.WordRange("off/on", AutolootSystem.binaryWords))
						.HandleWith(this.cmdSound)
					.EndSubCommand();

			this.harmony = new Harmony(harmonyId);
			harmony.PatchAll(Assembly.GetExecutingAssembly());
		}

		protected TextCommandResult cmdLog(TextCommandCallingArgs args) {
			string value = args[0] as string;

			string message;
			switch(value) {
				case AutolootSystem.wordFalse: {
					AutolootSystem.config.log = false;
					message = "Chat logging disabled.";
					break;
				}
				case AutolootSystem.wordTrue: {
					AutolootSystem.config.log = true;
					message = "Chat logging enabled.";
					break;
				}
				default: {
					return TextCommandResult.Error("");
				}
			}

			AutolootSystem.capi.StoreModConfig<AutolootConfig>(AutolootSystem.config, configFilename);
			return TextCommandResult.Success(print(message));
		}

		protected TextCommandResult cmdSound(TextCommandCallingArgs args) {
			string value = args[0] as string;

			string message;
			switch(value) {
				case AutolootSystem.wordFalse: {
					AutolootSystem.config.sound = false;
					message = "Sound effect disabled.";
					break;
				}
				case AutolootSystem.wordTrue: {
					AutolootSystem.config.sound = true;
					message = "Sound effect enabled.";
					break;
				}
				default: {
					return TextCommandResult.Error("");
				}
			}

			AutolootSystem.capi.StoreModConfig<AutolootConfig>(AutolootSystem.config, configFilename);
			return TextCommandResult.Success(print(message));
		}

		public override void Dispose() {
			base.Dispose();

			if(harmony != null) {
				this.harmony.UnpatchAll(harmonyId);
			}
		}

		public static void Process(GuiDialogCreatureContents dialog, InventoryGeneric inventory, EntityAgent entity) {
			if(AutolootSystem.Check(inventory)) {
				// Loot now.
				AutolootSystem.Loot(dialog, inventory, entity.EntityId, true);
			} else {
				// Wait for update.
				AutolootSystem.dialog = dialog;
				AutolootSystem.inventory = inventory;
				AutolootSystem.entityId = entity.EntityId;
				AutolootSystem.waiting = true;

				entity.WatchedAttributes.OnModified.Add(new TreeModifiedListener() { path = "harvestableInv", listener = onUpdate });
			}
		}

		protected static void onUpdate() {
			if(!AutolootSystem.waiting) {
				return;
			}

			if(AutolootSystem.Check(AutolootSystem.inventory)) {
				AutolootSystem.Loot(AutolootSystem.dialog, AutolootSystem.inventory, AutolootSystem.entityId, false);
			} else {
				AutolootSystem.dialog.TryClose();
			}

			AutolootSystem.dialog = null;
			AutolootSystem.inventory = null;
			AutolootSystem.entityId = -1;
			AutolootSystem.waiting = false;
		}

		public static bool Check(InventoryGeneric inventory) {
			if(inventory == null) {
				return false;
			}

			for(int i = 0; i < inventory.Count; i ++) {
				if(!inventory[i].Empty) {
					return true;
				}
			}
			return false;
		}

		public static void Loot(GuiDialogCreatureContents dialog, InventoryGeneric inventory, long entityId, bool queue) {
			if(dialog == null
					|| inventory == null) {
				return;
			}

			if(AutolootSystem.config.log) {
				AutolootSystem.sb.Append("+ ");
			}

			bool looted = false;
			for(int i = 0; i < inventory.Count; i ++) {
				if(inventory[i].Empty) {
					continue;
				}

				string name = inventory[i].GetStackName();
				int quantity = inventory[i].StackSize;

				ItemStackMoveOperation op = new ItemStackMoveOperation(AutolootSystem.capi.World, EnumMouseButton.Left, EnumModifierKey.SHIFT, EnumMergePriority.AutoMerge, inventory[i].StackSize);
				op.ActingPlayer = AutolootSystem.capi.World.Player;

				var packet = inventory.ActivateSlot(i, inventory[i], ref op);
				AutolootSystem.capi.Network.SendEntityPacket(entityId, packet);

				if(inventory[i].StackSize < quantity) {
					// Look taken.
					if(AutolootSystem.config.log) {
						if(looted) {
							AutolootSystem.sb.Append(", ");
						}

						AutolootSystem.sb.Append(name);
						int taken = quantity - inventory[i].StackSize;
						if(taken > 1) {
							AutolootSystem.sb.Append(" ×");
							AutolootSystem.sb.Append(taken);
						}
					}

					looted = true;
				}
			}

			if(looted) {
				if(AutolootSystem.config.log) {
					AutolootSystem.capi.ShowChatMessage(AutolootSystem.sb.ToString());
					AutolootSystem.sb.Clear();
				}

				if(AutolootSystem.config.sound) {
					AutolootSystem.capi.World.PlaySoundFor(AutolootSystem.sfx, AutolootSystem.capi.World.Player, false);
				}
			}

			if(!AutolootSystem.Check(inventory)) {
				// Close dialog.
				if(queue) {
					AutolootSystem.capi.Event.EnqueueMainThreadTask(() => dialog.TryClose(), "closedlg");
				} else {
					dialog.TryClose();
				}
			}
		}

		protected static string print(string text) {
			return "[Autoloot] " + text;
		}
	}

	[HarmonyPatch(typeof(GuiDialogCreatureContents), nameof(GuiDialogCreatureContents.OnGuiOpened))]
	public class AL_GDCC_OnGuiOpened {
		static void Postfix(GuiDialogCreatureContents __instance) {
			if(AutolootSystem.capi.Side != EnumAppSide.Client) {
				return;
			}

			Traverse self = Traverse.Create(__instance);
			InventoryGeneric inv = self.Field("inv").GetValue() as InventoryGeneric;
			EntityAgent owningEntity = self.Field("owningEntity").GetValue() as EntityAgent;

			IAttribute harvest = owningEntity.WatchedAttributes["harvestableInv"];
			if(harvest != null) {
				// Already loaded.
				if(!AutolootSystem.Check(inv)) {
					// Empty, close immediately.
					AutolootSystem.capi.Event.EnqueueMainThreadTask(() => __instance.TryClose(), "closedlg");
					return;
				}
			}

			AutolootSystem.Process(__instance, inv, owningEntity);
		}
	}
}